from keg_elements.db.columns import DBEnum


class AttemptType(DBEnum):
    forgot = "Forgot Password"
    login = "Login"
    reset = "Password Reset"

    @classmethod
    def db_name(cls):
        return "ka_attempt_types"



import arrow
import flask
import flask_login
import jinja2
import sqlalchemy as sa
from blazeutils import tolist
from keg.db import db
from keg.signals import db_init_post, init_complete
from webgrid.renderers import render_html_attributes

import keg_auth.cli
from keg_auth import model
from keg_auth.libs.authenticators import (
    DefaultPasswordPolicy,
    KegAuthenticator,
    OAuthAuthenticator,
)

DEFAULT_CRYPTO_SCHEMES = ('bcrypt', 'pbkdf2_sha256',)


class AuthManager(object):
    """Set up an auth management extension

    Main manager for keg-auth authentication/authorization functions, and provides a central
    location and handle on the flask app to access CLI setup, navigation, authenticators, etc.

    :param mail_manager: AuthMailManager instance used for mail functions. Can be None.
    :param blueprint: name to use for the blueprint containing auth views
    :param endpoints: dict of overrides to auth view endpoints
    :param cli_group_name: name of the CLI group under which auth commands appear
    :param grid_cls: webgrid class to serve as a base class to auth CRUD grids
    :param login_authenticator: login authenticator class used by login view
        default: KegAuthenticator
    :param request_loaders: registered loaders used for loading a user at request time from
        information not contained in the session (e.g. with an authorization header token).
        Can be scalar or an iterable
    :param permissions: permission strings defined for the app, which will be synced to the
        database on app init. Can be a single string or an iterable
    :param entity_registry: EntityRegistry instance on which User, Group, etc. are registered
    :param password_policy_cls: A PasswordPolicy class to check password requirements in
        forms and CLI
    """
    endpoints = {
        'forgot-password': '{blueprint}.forgot-password',
        'reset-password': '{blueprint}.reset-password',
        'change-password': '{blueprint}.change-password',
        'login': '{blueprint}.login',
        'logout': '{blueprint}.logout',
        'after-login': '{blueprint}.after-login',
        'after-logout': '{blueprint}.login',
        'after-forgot': '{blueprint}.login',
        'after-reset': '{blueprint}.login',
        'verify-account': '{blueprint}.verify-account',
        'after-verify-account': '{blueprint}.login',
        'oauth-login': '{blueprint}.oauth-login',
        'oauth-authorize': '{blueprint}.oauth-authorize',
    }
    cli_group_name = 'auth'

    def __init__(self, mail_manager=None, blueprint='auth', endpoints=None,
                 cli_group_name=None, grid_cls=None, login_authenticator=KegAuthenticator,
                 request_loaders=None, permissions=None, entity_registry=None,
                 oauth_authenticator=OAuthAuthenticator,
                 password_policy_cls=DefaultPasswordPolicy):
        self.mail_manager = mail_manager
        self.blueprint_name = blueprint
        self.entity_registry = entity_registry
        self.password_policy_cls = password_policy_cls
        self.endpoints = self.endpoints.copy()
        if endpoints:
            self.endpoints.update(endpoints)
        self.cli_group_name = cli_group_name or self.cli_group_name
        self.cli_group = None
        self.grid_cls = grid_cls
        self.login_authenticator_cls = login_authenticator
        self.oauth_authenticator_cls = oauth_authenticator
        self.request_loader_cls = tolist(request_loaders or [])
        self.request_loaders = dict()
        self.menus = dict()
        self.permissions = tolist(permissions or [])
        self._model_initialized = False
        self._loaders_initialized = False
        self._signal_handlers = []

    def init_app(self, app):
        """Inits KegAuth as a flask extension on the given app."""
        self.init_model(app)
        self.init_config(app)
        self.init_managers(app)
        self.init_cli(app)
        self.init_jinja(app)
        self.init_loaders(app)
        self.init_permissions(app)
        self.init_custom_csrf(app)

    def init_config(self, app):
        """Provide app config defaults for crypto, mail, logins, etc."""
        _cc_kwargs = dict(schemes=DEFAULT_CRYPTO_SCHEMES, deprecated='auto')
        app.config.setdefault('PASSLIB_CRYPTCONTEXT_KWARGS', _cc_kwargs)

        # config flag controls email ops such as sending verification emails, etc.
        # Note: model mixin must be in place for email
        app.config.setdefault('KEGAUTH_EMAIL_OPS_ENABLED', self.mail_manager is not None)

        site_name = app.config.get('SITE_NAME', 'UNKNOWN')
        app.config.setdefault('KEGAUTH_EMAIL_SITE_NAME', site_name)

        site_abbr = app.config.get('SITE_ABBR', site_name)
        app.config.setdefault('KEGAUTH_EMAIL_SITE_ABBR', site_abbr)

        app.config.setdefault('KEGAUTH_CRUD_INCLUDE_TITLE', True)
        app.config.setdefault('KEGAUTH_TEMPLATE_TITLE_VAR', 'page_title')
        app.config.setdefault('KEGAUTH_TOKEN_EXPIRE_MINS', 60 * 4)
        app.config.setdefault('KEGAUTH_LOGOUT_CLEAR_SESSION', True)

        app.config.setdefault('KEGAUTH_CLI_USER_ARGS', ['email'])

        # HTTP methods to ignore during auth checks. This can be useful for excluding
        # methods like OPTIONS during front-end API requests, for CORS compatibility.
        app.config.setdefault('KEGAUTH_HTTP_METHODS_EXCLUDED', [])

        # Set the login target for the redirect authenticator
        app.config.setdefault('KEGAUTH_REDIRECT_LOGIN_TARGET', None)

        # OAuth profiles
        app.config.setdefault('KEGAUTH_OAUTH_PROFILES', [])

        # Attempt lockout parameters.
        # - Enabled: default True, turns on attempt limits and requires the attempt entity.
        # - Limit: maximum number of attempts within the timespan.
        # - Timespan: number of seconds in which the limit can be reached.
        # - Lockout: number of seconds until an attempt can be made after the limit is reached.
        # - KEGAUTH_ATTEMPT_IP_LIMIT: base locking on IP address as well as input
        app.config.setdefault('KEGAUTH_ATTEMPT_LIMIT_ENABLED', True)
        app.config.setdefault('KEGAUTH_ATTEMPT_LIMIT', 15)
        app.config.setdefault('KEGAUTH_ATTEMPT_TIMESPAN', 600)  # 10 minutes
        app.config.setdefault('KEGAUTH_ATTEMPT_LOCKOUT', 3600)  # 1 hour
        app.config.setdefault('KEGAUTH_ATTEMPT_IP_LIMIT', True)
        # Lockout parameters may also be set for specific views (login, forgot, reset). If these
        # specific values are not set, the generic ATTEMPT values will be used.
        # app.config.setdefault('KEGAUTH_LOGIN_ATTEMPT_LIMIT', 3)
        # app.config.setdefault('KEGAUTH_LOGIN_ATTEMPT_TIMESPAN', 3600)  # 1 hour
        # app.config.setdefault('KEGAUTH_LOGIN_ATTEMPT_LOCKOUT', 3600)  # 1 hour
        # app.config.setdefault('KEGAUTH_FORGOT_ATTEMPT_LIMIT', 5)
        # app.config.setdefault('KEGAUTH_FORGOT_ATTEMPT_TIMESPAN', 3600)  # 1 hour
        # app.config.setdefault('KEGAUTH_FORGOT_ATTEMPT_LOCKOUT', 3600)  # 1 hour
        # app.config.setdefault('KEGAUTH_RESET_ATTEMPT_LIMIT', 1)
        # app.config.setdefault('KEGAUTH_RESET_ATTEMPT_TIMESPAN', 86400)  # 24 hours
        # app.config.setdefault('KEGAUTH_RESET_ATTEMPT_LOCKOUT', 86400)  # 24 hours

    def init_cli(self, app):
        """Add a CLI group for auth."""
        keg_auth.cli.add_cli_to_app(app, self.cli_group_name,
                                    user_args=app.config.get('KEGAUTH_CLI_USER_ARGS'))

    def init_jinja(self, app):
        """Set up app jinja loader to use keg-auth templates, select2, etc."""
        loader = jinja2.ChoiceLoader([
            app.jinja_loader,
            jinja2.PackageLoader('keg_auth', 'templates'),
            # Get access to form generation templates from Keg Elements.
            jinja2.PackageLoader('keg_elements', 'templates'),
        ])
        app.jinja_loader = loader
        app.context_processor(lambda: {
            'auth_manager': self,
            'use_select2': app.config.get('KEGAUTH_USE_SELECT2'),
        })
        app.jinja_env.filters['html_attributes'] = app.jinja_env.filters.get(
            'html_attributes', render_html_attributes
        )

    def init_model(self, app):
        """Set up the entity registry for all auth objects."""
        if not self._model_initialized:
            model.initialize_mappings(registry=self.entity_registry)
            model.initialize_events(registry=self.entity_registry)
            self._model_initialized = True

    def init_managers(self, app):
        """Place this extension on the app for reference, and onfigure flask-login."""
        app.auth_manager = self

        app.login_manager = login_manager = flask_login.LoginManager()
        login_manager.user_loader(self.user_loader)
        if app.testing:
            login_manager.request_loader(self.test_request_loader)
        login_manager.login_view = self.endpoint('login')
        login_manager.init_app(app)

    def init_loaders(self, app):
        """Initialize user session loaders."""
        if self._loaders_initialized:
            return

        self.login_authenticator = self.login_authenticator_cls(app)

        if app.config.get('KEGAUTH_OAUTH_PROFILES'):
            self.oauth_authenticator = OAuthAuthenticator(app)

        for loader_cls in self.request_loader_cls:
            self.request_loaders[loader_cls.get_identifier()] = loader_cls(app)

        self._loaders_initialized = True

    def init_permissions(self, app):
        """Configure database with the defined set of permissions.

        Synchronizes permission records in the database with those defined in the app. Ensures
        the sync method is called in the proper place during test runs, when the database is not
        fully available and set up at extension-loading time.
        """
        from keg_auth.model.entity_registry import RegistryError

        if not self.entity_registry:
            return

        try:
            Permission = self.entity_registry.permission_cls
        except RegistryError:
            return

        # The tricky thing here is that the db may not be ready. Normal app startup should
        # expect it at this point, but test setup may not have initialized tables by now.
        # So, connect it to the test signal, then try to call it, and trap the exception
        def sync_permissions(app):
            with app.app_context():
                db_permissions = db.session.query(Permission).all()

                # sync permission presence
                desired_tokens = set(
                    tolist(perm)[0] for perm in app.auth_manager.permissions
                )
                current_tokens = {
                    permission.token for permission in db_permissions
                }
                for permission in desired_tokens - current_tokens:
                    db_permissions.append(Permission.add(token=permission, _commit=False))
                for permission in current_tokens - desired_tokens:
                    Permission.query.filter_by(token=permission).delete()

                # sync permission description
                permission_descriptions = dict(
                    [perm for perm in app.auth_manager.permissions
                     if isinstance(perm, (list, tuple))]
                )
                for db_permission in db_permissions:
                    if (
                        db_permission.token in permission_descriptions
                        and db_permission.description
                            != permission_descriptions[db_permission.token]
                    ):
                        db_permission.description = permission_descriptions[db_permission.token]
                    elif (
                        db_permission.token not in permission_descriptions
                        and db_permission.description
                    ):
                        db_permission.description = None

                try:
                    db.session.commit()
                except sa.exc.IntegrityError as exc:
                    # We have a possible race condition here, in that if another app process starts
                    # while we are computing permission sync, we could get an integrity error.
                    # Check that it is a unique exception, but note that we're not able to check
                    # the constraint name here (too many assumptions to be made).
                    from keg_elements.db.utils import validate_unique_exc
                    if not validate_unique_exc(exc):
                        raise

        # store the connected method somewhere, so we don't lose it with current function scope
        self._sync_permissions = db_init_post.connect(sync_permissions)

        try:
            # During normal app startup with models present, this should work just fine. However,
            # if we are running tests where the model is cleaned/restored during test setup,
            # that process completes after this step. So, we need to trap the ensuing exception,
            # and let the testing signal do the setup.
            # syncing permissions during testing was causing some db session issues.
            if not app.testing:
                sync_permissions(app)
        except sa.exc.ProgrammingError as exc:
            if 'permissions' not in str(exc):
                raise

    def init_custom_csrf(self, app):
        @init_complete.connect
        def handle_custom_csrf(app):
            if 'csrf' not in app.extensions:
                return
            for endpoint_key in self.endpoints.keys():
                endpoint = self.endpoint(endpoint_key)
                view_obj = app.view_functions.get(endpoint)
                if (
                    not hasattr(view_obj, 'view_class')
                    or not hasattr(view_obj.view_class, 'auth_manager_key')
                    or not hasattr(view_obj.view_class, 'responder_key')
                ):
                    # only support auth-responded class views for this right now
                    continue
                dest = f"{view_obj.__module__}.{view_obj.__name__}"
                authenticator_cls = getattr(self, f'{view_obj.view_class.auth_manager_key}_cls')
                responder_cls = authenticator_cls.responder_cls.get(
                    view_obj.view_class.responder_key
                )
                if getattr(responder_cls, '_csrf_custom_handling', False):
                    app.extensions['csrf'].exempt(dest)

        # persist a reference to this method for the signal to call
        self._signal_handlers.append(handle_custom_csrf)

    def validate_permission_set(self, permissions):
        defined_tokens = set(
            tolist(perm)[0] for perm in self.permissions
        )
        undefined_tokens = set(tolist(permissions)) - defined_tokens
        if undefined_tokens:
            raise Exception(f'permission(s) {undefined_tokens} not specified in the auth manager')

    def add_navigation_menu(self, name, menu):
        """Create a navigation menu that may be referenced with the given name."""
        self.menus[name] = menu

    def endpoint(self, ident):
        """Return an auth endpoint on the configured blueprint."""
        return self.endpoints[ident].format(blueprint=self.blueprint_name)

    def url_for(self, ident, **kwargs):
        """Generate the URL for the endpoint identified by `ident`."""
        return flask.url_for(self.endpoint(ident), **kwargs)

    def user_loader(self, session_key):
        """Fetch a user record via session key."""
        user_class = self.entity_registry.user_cls
        return user_class.get_by(session_key=str(session_key))

    def user_by_id(self, user_id):
        """Fetch a user record via ID."""
        user_class = self.entity_registry.user_cls
        return user_class.get_by(id=user_id)

    def test_request_loader(self, request):
        """ Load a user from a request when testing. This gives a nice API for test clients to
            be logged in, rather than expecting all tests to set up an actual session.

            See `keg_auth.testing.AuthTestApp` for a webtest wrapper using this loader.
        """
        session_key = request.environ.get('TEST_USER_ID')
        if session_key is None:
            return
        return self.user_loader(session_key)

    def create_user_cli(self, extra_args=None, **kwargs):
        """ A thin layer between the cli and `create_user()` to transform the cli args
            into what the User entity expects for fields.

            For example, if you had a required `name` field on your User entity, then you could do
            something like::

                $ yourkegapp auth create-user john.smith@example.com "John Smith"

            Then this method would get overriden in a sub-class like:

                def create_user_cli(self, email, extra_args):
                    user_kwargs = dict(email=email, name=extra_args[0])
                    return self.create_user(user_kwargs)
        """
        # By default, we assume no extra arguments are used
        user_kwargs = kwargs
        return self.create_user(user_kwargs)

    def create_user(self, user_kwargs, _commit=True):
        """Create a new user record and optionally persist to the database.

        :param user_kwargs: dict of values to construct the User record. Special arg is
            `mail_enabled`, which will be popped out.
        :param _commit: option for persisting record to database. Default True.
        """
        mail_enabled = user_kwargs.pop('mail_enabled', True)
        from passlib.pwd import genword
        user_kwargs.setdefault('password', genword(entropy='secure'))
        user_class = self.entity_registry.user_cls
        user = user_class(**user_kwargs)
        db.session.add(user)
        db.session.flush()

        # generate the token AFTER flush, because the token may depend on things like timestamps
        # which may not be available earlier
        user.token_generate()

        if (
            mail_enabled
            and self.mail_manager
            and not self.login_authenticator.is_domain_excluded(user.username)
        ):
            self.mail_manager.send_new_user(user)

        # use add + commit here instead of user_class.add() above so the user isn't actually
        # committed if mail isn't sent.
        if _commit:
            db.session.commit()
        return user

    def get_request_loader(self, identifier):
        """Returns a registered request loader, keyed by its identifier."""
        return self.request_loaders.get(identifier)

    def resend_verification_email(self, user_id):
        """Generate a fresh token and send the account verification email."""
        user = self.user_by_id(user_id)
        if not self.mail_manager:
            raise Exception("Tried to resend verification email, but email is not setup.")
        # ensure the user object has a fresh token
        user.token_generate()
        self.mail_manager.send_new_user(user)


# ensure that any manager-attached menus are reset for auth requirements on login/logout
def refresh_session_menus(app, user):
    for menu in app.auth_manager.menus.values():
        menu.clear_authorization(user.get_id())


def update_last_login(app, user):
    user.last_login_utc = arrow.utcnow()
    db.session.commit()


def on_login(app, user):
    refresh_session_menus(app, user)
    update_last_login(app, user)


def clear_session(app, user):
    if app.config.get('KEGAUTH_LOGOUT_CLEAR_SESSION'):
        flask.session.clear()


flask_login.signals.user_logged_in.connect(on_login)
flask_login.signals.user_logged_out.connect(refresh_session_menus)
flask_login.signals.user_logged_out.connect(clear_session)



from unittest import mock

import flask
import freezegun
import arrow
import flask_webtest
from keg.db import db
import pytest
import sqlalchemy as sa
from webgrid.extensions import RequestArgsLoader, RequestFormLoader
from werkzeug.datastructures import MultiDict
from keg.testing import ContextManager
from keg_auth_ta.app import KegAuthTestApp, mail_ext
from keg_auth.libs.authenticators import OAuthAuthenticator, RedirectAuthenticator
from keg_auth.libs.decorators import requires_user
from keg_auth.testing import AuthTests, AuthTestApp, ViewTestBase
from keg_auth.tests.utils import oauth_profile

from keg_auth_ta.model import entities as ents
import flask_login
from .utils import listen_to

from keg_auth import get_current_user


@pytest.fixture(scope='function')
def app_context():
    with ContextManager.get_for(KegAuthTestApp).app.app_context():
        yield


class TestAuthIntegration(AuthTests):
    user_ent = ents.User


class TestViews(object):
    """
        Basic functionality is tested through AuthTests.  The tests in this class cover
        functionality that is specific to the default implementation but might fail depending on
        what customization is made.
    """

    @classmethod
    def setup_class(cls):
        ents.Permission.delete_cascaded()
        cls.perm1 = ents.Permission.fake(token='permission1')
        cls.perm2 = ents.Permission.fake(token='permission2')
        cls.perm_auth = ents.Permission.fake(token='auth-manage')

    def setup_method(self):
        ents.User.delete_cascaded()

    def test_class_decorator_throws_exception(self):
        with pytest.raises(TypeError):
            @requires_user
            class NotKegOrFlask:
                pass

    def test_home(self):
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/')
        assert 'Keg Auth Demo' in resp
        assert '/login' in resp
        assert '/logout' not in resp

        user = ents.User.fake()
        with client.session_transaction() as sess:
            sess['_user_id'] = user.session_key

        resp = client.get('/')
        assert '/login' not in resp
        assert '/logout' in resp

    def test_auth_base_view(self):
        ents.User.fake(email='foo@bar.com', password='pass',
                       permissions=[self.perm1, self.perm2])

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/secret2', status=302)

        resp = resp.follow()
        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'pass'
        resp = resp.form.submit(status=302)
        assert resp.flashes == [('success', 'Login successful.')]

        # Now that we have logged in, we should be able to get to the page.
        resp = client.get('/secret2', status=200)
        assert resp.text == 'secret2'

    def test_rendered_navigation(self):
        ents.User.fake(email='foo@bar.com', password='pass',
                       permissions=[self.perm_auth])

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/')
        doc = resp.pyquery
        assert doc.find('div#navigation a[href="/"]')
        assert not doc.find('div#navigation a[href="/users"]')
        assert not doc.find('div#navigation a[href="/secret-nested"]')

        resp = client.get('/login')
        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'pass'
        resp = resp.form.submit(status=302)

        resp = client.get('/')
        doc = resp.pyquery
        assert doc.find('div#navigation a[href="/"]')
        assert doc.find('div#navigation a[href="/users"]')
        assert not doc.find('div#navigation a[href="/secret-nested"]')

        client.get('/logout')
        resp = client.get('/')
        doc = resp.pyquery
        assert doc.find('div#navigation a[href="/"]')
        assert not doc.find('div#navigation a[href="/users"]')
        assert not doc.find('div#navigation a[href="/secret-nested"]')

    def test_navigation_group(self):
        user = ents.User.fake(permissions=[self.perm_auth])
        client = AuthTestApp(flask.current_app, user=user)
        resp = client.get('/')
        nav_el = resp.pyquery('#navigation')
        assert nav_el('[href="#navgroup-auth"]').attr('aria-expanded') != 'true'
        assert len(nav_el('[aria-expanded="true"]')) == 0
        assert not nav_el('#navgroup-auth').has_class('in')
        assert nav_el('.nav-current').text() == 'Home'

        resp = client.get('/users')
        nav_el = resp.pyquery('#navigation')
        assert nav_el('[href="#navgroup-auth"]').attr('aria-expanded') == 'true'
        assert len(nav_el('[aria-expanded="true"]')) == 1
        assert nav_el('#navgroup-auth').has_class('in')

        with mock.patch.object(flask.current_app, 'template_globals', {'auto_expand_menu': False}):
            resp = client.get('/users')
            nav_el = resp.pyquery('#navigation')
            assert nav_el('[href="#navgroup-auth"]').attr('aria-expanded') != 'true'

    def test_navigation_group_class(self):
        user = ents.User.fake(permissions=[self.perm_auth, self.perm1, self.perm2])
        client = AuthTestApp(flask.current_app, user=user)
        resp = client.get('/')
        nav_el = resp.pyquery('#navigation')
        assert nav_el('.group-header.my-group-class').text() == 'Sub-Menu'

    def test_navigation_link_class(self):
        user = ents.User.fake(permissions=[self.perm_auth, self.perm1, self.perm2])
        client = AuthTestApp(flask.current_app, user=user)
        resp = client.get('/')
        nav_el = resp.pyquery('#navigation')
        assert nav_el('.my-link-class').text() == 'Secret View'

    def test_navigation_icon(self):
        user = ents.User.fake(permissions=[self.perm_auth])
        client = AuthTestApp(flask.current_app, user=user)
        resp = client.get('/')
        nav_el = resp.pyquery('#navigation')
        assert nav_el('[href="#navgroup-auth"]')('i.fas.fa-bomb')
        assert len(nav_el('i.fas.fa-bomb')) == 1
        assert '<i class="fas fa-ad"/>User Manage 3' in nav_el.html()
        assert len(nav_el('i.fas.fa-ad')) == 1

    def test_authenticated_client(self):
        user = ents.User.fake()
        client = AuthTestApp(flask.current_app, user=user)
        resp = client.get('/secret1', status=200)
        assert resp.text == 'secret1'

        resp = client.get('/secret1-class', status=200)
        assert resp.text == 'secret1-class'

    def test_unauthenticated_client(self):
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/secret1', status=302)
        assert '/login' in resp.location
        # decorated class has its own authentication failure handler
        client.get('/secret1-class', status=405)

    @mock.patch('flask.current_app.auth_manager.request_loaders', {})
    def test_unauthenticated_client_no_request_loaders(self):
        client = flask_webtest.TestApp(flask.current_app)
        client.get('/secret1', status=302)

    @freezegun.freeze_time("2018-10-01 15:00:00")
    def test_login_user_sets_last_login_and_invalidates_token(self):
        u = ents.User.fake(email='foo@bar.com', password='pass', last_login_utc=None)
        user_id = u.id
        token = u.token_generate()
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/login', status=200)

        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'pass'
        with listen_to(flask_login.user_logged_in) as listener:
            resp = resp.form.submit()
        assert len(listener.heard) == 1

        # Can't just assert against the listener args, because the user instance is
        # from another app context
        db.session.remove()
        db.session.add(listener.heard[0][1]['user'])
        assert listener.heard[0][0][0].__class__.__name__ == flask.current_app.__class__.__name__
        assert listener.heard[0][1]['user'].id == user_id

        assert resp.status_code == 302, resp.html
        db.session.remove()
        u = ents.User.get_by(email="foo@bar.com")
        assert u.last_login_utc == arrow.utcnow()
        assert not u.token_verify(token)

    def test_login_field_success_next_parameter(self):
        ents.User.fake(email='foo@bar.com', password='pass')

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/secret1', status=302).follow()

        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'pass'
        resp = resp.form.submit()

        assert resp.status_code == 302, resp.html
        assert resp.headers['Location'] == '/secret1'
        assert resp.flashes == [('success', 'Login successful.')]

    def test_login_field_success_next_session(self):
        ents.User.fake(email='foo@bar.com', password='pass')

        with mock.patch.dict(flask.current_app.config, {'USE_SESSION_FOR_NEXT': True}):
            client = flask_webtest.TestApp(flask.current_app)
            resp = client.get('/secret1', status=302).follow()

            resp.form['login_id'] = 'foo@bar.com'
            resp.form['password'] = 'pass'
            resp = resp.form.submit()

        assert resp.status_code == 302, resp.html
        assert resp.headers['Location'] == '/secret1'
        assert resp.flashes == [('success', 'Login successful.')]

    def test_authenticated_request(self):
        user = ents.User.fake(permissions=[self.perm1, self.perm2])
        client = AuthTestApp(flask.current_app)

        resp = client.get('/secret2', status=200, user=user)
        assert resp.text == 'secret2'

        # User should only stick around for a single request
        client.get('/secret2', status=302)

        # test all HTTP methods
        assert client.post('/secret2', status=200, user=user).text == 'secret2 post'
        assert client.put('/secret2', status=200, user=user).text == 'secret2 put'
        assert client.patch('/secret2', status=200, user=user).text == 'secret2 patch'
        assert client.delete('/secret2', status=200, user=user).text == 'secret2 delete'
        assert client.options('/secret2', status=200, user=user).text == 'secret2 options'
        assert client.head('/secret2', status=200, user=user)
        assert client.post_json('/secret2', status=200, user=user).text == 'secret2 post'
        assert client.put_json('/secret2', status=200, user=user).text == 'secret2 put'
        assert client.patch_json('/secret2', status=200, user=user).text == 'secret2 patch'
        assert client.delete_json('/secret2', status=200, user=user).text == 'secret2 delete'

    def test_login_template(self):
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/login')
        doc = resp.pyquery
        assert doc('title').text() == 'Log In | Keg Auth Demo'
        assert doc('h1').text() == 'Log In'
        assert doc('button').text() == 'Log In'
        assert doc('div#page-content a').text() == 'I forgot my password'
        assert doc('div#page-content a').attr('href') == '/forgot-password'

    @mock.patch.dict('flask.current_app.config', {'KEGAUTH_EMAIL_OPS_ENABLED': False})
    def test_login_template_no_mail(self):
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/login')
        doc = resp.pyquery
        assert doc('title').text() == 'Log In | Keg Auth Demo'
        assert doc('h1').text() == 'Log In'
        assert doc('button').text() == 'Log In'
        assert not doc('div#page-content a')

    def test_login_head(self):
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.head(
            flask.url_for(flask.current_app.auth_manager.endpoint('login')),
            status=405,
        )
        assert resp.headers['Allow'] == 'GET, POST'

    def test_forgot_pw_template(self):
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/forgot-password')
        doc = resp.pyquery
        assert doc('title').text() == 'Initiate Password Reset | Keg Auth Demo'
        assert doc('h1').text() == 'Initiate Password Reset'
        assert doc('button').text() == 'Send Reset Email'
        assert doc('div#page-content a').text() == 'Cancel'
        assert doc('div#page-content a').attr('href') == '/login'

    @mock.patch('keg_auth.views.flask.current_app.auth_manager.mail_manager.send_reset_password',
                autospec=True, spec_set=True)
    def test_forget_pw_actions(self, m_send_reset_password):
        user = ents.User.fake(email='foo@bar.com')

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/forgot-password')
        resp.form['email'] = 'foo@bar.com'
        resp = resp.form.submit(status=302)

        # Email should be sent. Can't just assert called with the user instance,
        # because the client is in a different app context, so has a different
        # session and user instance.
        m_send_reset_password.assert_called_once()
        assert m_send_reset_password.call_args[0][0].id == user.id

    @mock.patch('flask.current_app.auth_manager.mail_manager', None)
    def test_forget_pw_actions_mail_disabled(self):
        client = flask_webtest.TestApp(flask.current_app)
        client.get('/forgot-password', status=404)

    def test_reset_pw_template(self):
        user = ents.User.fake()
        user.token_generate()
        url = flask.current_app.auth_manager.mail_manager.reset_password_url(user)

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(url)
        doc = resp.pyquery
        assert doc('title').text() == 'Complete Password Reset | Keg Auth Demo'
        assert doc('h1').text() == 'Complete Password Reset'
        assert doc('button').text() == 'Change Password'
        assert doc('div#page-content a').text() == 'Cancel'
        assert doc('div#page-content a').attr('href') == '/login'

    def test_reset_pw_head(self):
        user = ents.User.fake()
        token = user.token_generate()

        client = flask_webtest.TestApp(flask.current_app)
        client.head('/reset-password/{}/{}'.format(user.id, token), status=405)

    def test_reset_pw_actions(self):
        user = ents.User.fake()
        token = user.token_generate()

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/reset-password/{}/{}'.format(user.id, token))
        resp.form['password'] = resp.form['confirm'] = 'foobarbaz'
        resp = resp.form.submit(status=302)

        # Make sure db updates got committed
        db.session.expire(user)
        assert user.password == 'foobarbaz'

    @mock.patch('flask.current_app.auth_manager.mail_manager', None)
    def test_reset_pw_actions_mail_disabled(self):
        user = ents.User.fake()
        token = user.token_generate()

        client = flask_webtest.TestApp(flask.current_app)
        client.get('/reset-password/{}/{}'.format(user.id, token), status=404)

    def test_verify_account_template(self):
        user = ents.User.fake()
        user.token_generate()
        url = flask.current_app.auth_manager.mail_manager.verify_account_url(user)

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(url)
        doc = resp.pyquery
        assert doc('title').text() == 'Verify Account & Set Password | Keg Auth Demo'
        assert doc('h1').text() == 'Verify Account & Set Password'
        assert doc('button').text() == 'Verify & Set Password'
        assert doc('div#page-content a').text() == 'Cancel'
        assert doc('div#page-content a').attr('href') == '/login'


class TestPermissionsRequired:
    @classmethod
    def setup_class(cls):
        ents.Permission.delete_cascaded()
        cls.perm1 = ents.Permission.fake(token='permission1')
        cls.perm2 = ents.Permission.fake(token='permission2')
        cls.perm3 = ents.Permission.fake(token='permission3')

    def test_custom_authentication_failure(self):
        allowed = ents.User.fake()
        client = AuthTestApp(flask.current_app, user=allowed)
        resp = client.get('/custom-auth-failure', status=200)
        assert resp.text == 'custom-auth-failure'

        client = flask_webtest.TestApp(flask.current_app)
        client.get('/custom-auth-failure', status=400)

    def test_custom_authorization_failure(self):
        allowed = ents.User.fake(permissions=[self.perm1])
        disallowed = ents.User.fake()
        client = AuthTestApp(flask.current_app, user=allowed)
        resp = client.get('/custom-perm-failure', status=200)
        assert resp.text == 'custom-perm-failure'

        client = AuthTestApp(flask.current_app, user=disallowed)
        client.get('/custom-perm-failure', status=400)

    def test_method_level(self):
        allowed = ents.User.fake(permissions=[self.perm1, self.perm2])
        disallowed = ents.User.fake(permissions=[self.perm1])

        client = AuthTestApp(flask.current_app, user=allowed)
        resp = client.get('/secret2', status=200)
        assert resp.text == 'secret2'

        client = AuthTestApp(flask.current_app, user=disallowed)
        client.post('/secret2', {}, status=403)

    @pytest.mark.parametrize('endpoint', ['secret3', 'secret-flask'])
    def test_class_level(self, endpoint):
        allowed = ents.User.fake(permissions=[self.perm1, self.perm2])
        disallowed = ents.User.fake(permissions=[self.perm1])

        client = AuthTestApp(flask.current_app, user=allowed)
        resp = client.get('/{}'.format(endpoint), status=200)
        assert resp.text == endpoint

        client = AuthTestApp(flask.current_app, user=disallowed)
        client.get('/{}'.format(endpoint), {}, status=403)

        client = flask_webtest.TestApp(flask.current_app)
        client.get('/{}'.format(endpoint), status=302)

    @pytest.mark.parametrize('endpoint', ['secret3-sub', 'secret-flask-sub'])
    def test_class_level_inheritance(self, endpoint):
        allowed = ents.User.fake(permissions=[self.perm1, self.perm2])
        disallowed = ents.User.fake(permissions=[self.perm1])

        client = AuthTestApp(flask.current_app, user=allowed)
        resp = client.get('/{}'.format(endpoint), status=200)
        assert resp.text == endpoint

        client = AuthTestApp(flask.current_app, user=disallowed)
        client.get('/{}'.format(endpoint), {}, status=403)

        client = flask_webtest.TestApp(flask.current_app)
        client.get('/{}'.format(endpoint), status=302)

    @pytest.mark.parametrize('endpoint', ['secret4', 'secret-flask4'])
    def test_class_and_method_level_combined(self, endpoint):
        allowed = ents.User.fake(permissions=[self.perm1, self.perm2])
        disallowed1 = ents.User.fake(permissions=[self.perm1])
        disallowed2 = ents.User.fake(permissions=[self.perm2])

        client = AuthTestApp(flask.current_app, user=allowed)
        resp = client.get('/{}'.format(endpoint), status=200)
        assert resp.text == endpoint

        client = AuthTestApp(flask.current_app, user=disallowed1)
        client.get('/{}'.format(endpoint), {}, status=403)

        # missing the class-level permission requirement triggers the class's custom auth
        # failure handler
        client = AuthTestApp(flask.current_app, user=disallowed2)
        client.get('/{}'.format(endpoint), {}, status=405)

        client = flask_webtest.TestApp(flask.current_app)
        client.get('/{}'.format(endpoint), status=302)

    def test_nested_conditions(self):
        def check(perms, allowed):
            print(perms, allowed)
            target_status = 200 if allowed else 403
            user = ents.User.fake(permissions=perms)

            client = AuthTestApp(flask.current_app, user=user)
            resp = client.get('/secret-nested', status=target_status)
            if allowed:
                assert resp.text == 'secret_nested'

        for perms, allowed in (
            ((self.perm1, self.perm2), True),
            ((self.perm3,), True),
            ((self.perm1,), False),
            ((self.perm2,), False),
            ((self.perm1, self.perm2, self.perm3), True),
        ):
            check(perms, allowed)

    def test_nested_callable_conditions(self):
        def check(perms, email, allowed):
            print(perms, email, allowed)
            ents.User.delete_cascaded()
            target_status = 200 if allowed else 403
            user = ents.User.fake(permissions=perms, email=email)

            client = AuthTestApp(flask.current_app, user=user)
            resp = client.get('/secret-nested-callable', status=target_status)
            if allowed:
                assert resp.text == 'secret_nested_callable'

        for perms, email, allowed in (
            ((self.perm1,), 'snoopy@peanuts.com', True),
            ((self.perm2,), 'snoopy@peanuts.com', False),
            ((self.perm2,), 'foo@bar.baz', True),
        ):
            check(perms, email, allowed)

    def test_callable_conditions(self):
        def check(email, allowed):
            print(email, allowed)
            ents.User.delete_cascaded()
            target_status = 200 if allowed else 403
            user = ents.User.fake(email=email)

            client = AuthTestApp(flask.current_app, user=user)
            resp = client.get('/secret-callable', status=target_status)
            if allowed:
                assert resp.text == 'secret_callable'

        for email, allowed in (
            ('snoopy@peanuts.com', False),
            ('foo@bar.baz', True),
        ):
            check(email, allowed)

    def test_blueprint_method_level(self):
        allowed = ents.User.fake(permissions=[self.perm1])
        disallowed = ents.User.fake()

        client = AuthTestApp(flask.current_app, user=allowed)
        resp = client.get('/protected-method', status=200)
        assert resp.text == 'protected-method'

        client = AuthTestApp(flask.current_app, user=disallowed)
        client.get('/protected-method', status=403)

    def test_blueprint_class_level(self):
        allowed = ents.User.fake(permissions=[self.perm1])
        disallowed = ents.User.fake()

        client = AuthTestApp(flask.current_app, user=allowed)
        resp = client.get('/protected-class', status=200)
        assert resp.text == 'protected-class'
        resp = client.get('/protected-class2')
        assert resp.text == 'protected-class2'

        client = AuthTestApp(flask.current_app, user=disallowed)
        client.get('/protected-class', {}, status=403)
        client.get('/protected-class2')

        # blueprint has custom authentication failure hander
        client = flask_webtest.TestApp(flask.current_app)
        client.get('/protected-class', status=405)
        client.get('/protected-class2', status=302)

    def test_user_requirement_excludes_method(self):
        client = flask_webtest.TestApp(flask.current_app)
        # should have a 200 response
        client.options('/protected-class2')


class TestRequestLoaders(object):
    def test_token_auth_no_token(self):
        client = flask_webtest.TestApp(flask.current_app)
        client.get('/jwt-required', status=302)

    def test_token_auth_with_token(self):
        user = ents.User.fake()
        jwt_auth = flask.current_app.auth_manager.get_request_loader('jwt')
        token = jwt_auth.create_access_token(user)

        client = flask_webtest.TestApp(flask.current_app)
        client.set_authorization(('Bearer', token))
        resp = client.get('/jwt-required', status=200)
        assert resp.text == 'jwt-required'

    def test_per_endpoint_loader(self):
        ents.User.fake()

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/custom-loader-denied', status=302)
        assert resp.location.startswith('/login')
        resp = client.get('/custom-loader-letmein')
        assert resp.text == 'ok'


class TestUserCrud(ViewTestBase):
    permissions = 'auth-manage'

    def test_add(self):
        perm_approve = ents.Permission.fake()
        ents.Permission.fake()
        group_approve = ents.Group.fake()
        ents.Group.fake()
        bundle_approve = ents.Bundle.fake()
        ents.Bundle.fake()

        resp = self.client.get('/users/add')

        assert resp.pyquery('#page-content')('h1').eq(0).text() == 'Create User'
        assert resp.form['email'].value == ''
        assert 'is_superuser' not in resp.form.fields, resp.form.fields.keys()

        resp.form['email'] = 'abc@example.com'
        resp.form['permission_ids'] = [perm_approve.id]
        resp.form['group_ids'] = [group_approve.id]
        resp.form['bundle_ids'] = [bundle_approve.id]
        resp.form['disabled_utc'] = '2019-01-01'
        with mail_ext.record_messages() as outbox:
            resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/users')
        assert resp.flashes == [('success', 'Successfully created User')]
        assert len(outbox) == 1
        assert outbox[0].subject == '[KA Demo] User Welcome & Verification'
        user = self.user_ent.get_by(email='abc@example.com')
        assert user.disabled_utc == arrow.get('2019-01-01')
        assert user.is_enabled is True
        assert user.is_superuser is False
        assert user.permissions == [perm_approve]
        assert user.groups == [group_approve]
        assert user.bundles == [bundle_approve]

    def test_add_no_welcome(self):
        resp = self.client.get('/users/add')

        resp.form['email'] = 'abc3@example.com'
        resp.form['send_welcome'] = False
        with mail_ext.record_messages() as outbox:
            resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/users')
        assert resp.flashes == [('success', 'Successfully created User')]
        assert len(outbox) == 0
        assert self.user_ent.get_by(email='abc3@example.com')

    def test_resend_verification(self):
        self.current_user.is_verified = True
        self.current_user.permissions = ents.Permission.query.filter_by(token='auth-manage').all()
        for user in ents.User.query.filter(ents.User.email != self.current_user.email):
            ents.db.session.delete(user)
        user_edit = self.user_ent.fake(
            is_verified=False,
            email="foo1@bar.com",
            is_superuser=False,
            permissions=[]
        )
        resp = self.client.get('/users')
        form = resp.forms[1]

        # assert only one verification link because
        # only one unverified user
        assert 'user_id' not in resp.forms[0].fields
        assert len(resp.forms) == 2
        assert 'user_id' in form.fields
        assert form['user_id'].value == str(user_edit.id)
        with mail_ext.record_messages() as outbox:
            resp = form.submit()

        assert len(outbox) == 1
        assert outbox[0].subject == '[KA Demo] User Welcome & Verification'
        assert 'foo1@bar.com' in outbox[0].recipients
        assert resp.status_code == 302
        assert resp.location.endswith('/users')
        assert resp.flashes == [('success', 'Verification email has been sent')]

    def test_verification_column_appears_when_needed(self):
        resp = self.client.get('/users?op(username)=eq&v1(username)=' + self.current_user.email)
        assert resp.pyquery('.datagrid table.records thead th').eq(4).text() == 'Resend Verification' # noqa
        with mock.patch.dict('flask.current_app.config', {'KEGAUTH_EMAIL_OPS_ENABLED': False}):
            resp = self.client.get('/users?op(username)=eq&v1(username)=' + self.current_user.email)
            assert resp.pyquery('.datagrid table.records thead th').eq(4).text() == ''

    def test_add_and_token_is_correct(self):
        perm_approve = ents.Permission.fake()
        ents.Permission.fake()
        group_approve = ents.Group.fake()
        ents.Group.fake()
        bundle_approve = ents.Bundle.fake()
        ents.Bundle.fake()

        resp = self.client.get('/users/add')

        assert resp.form['email'].value == ''
        assert 'is_superuser' not in resp.form.fields, resp.form.fields.keys()

        resp.form['email'] = 'abc2@example.com'
        resp.form['permission_ids'] = [perm_approve.id]
        resp.form['group_ids'] = [group_approve.id]
        resp.form['bundle_ids'] = [bundle_approve.id]
        with mail_ext.record_messages() as outbox:
            resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/users')
        assert resp.flashes == [('success', 'Successfully created User')]
        assert len(outbox) == 1
        assert outbox[0].subject == '[KA Demo] User Welcome & Verification'
        assert 'abc2@example.com' in outbox[0].recipients
        user = self.user_ent.get_by(email='abc2@example.com')
        assert user.token_generate() in outbox[0].as_string()
        assert user.is_enabled is True
        assert user.is_superuser is False
        assert user.permissions == [perm_approve]
        assert user.groups == [group_approve]
        assert user.bundles == [bundle_approve]

    @mock.patch.dict('flask.current_app.config', {'KEGAUTH_EMAIL_OPS_ENABLED': False})
    def test_add_no_email(self):
        resp = self.client.get('/users/add')

        resp.form['email'] = 'foobar@baz.com'
        resp = resp.form.submit()

        assert resp.pyquery('#reset_password').siblings('.invalid-feedback').text() == \
            'This field is required.'
        resp.form['reset_password'] = 'bleh'
        resp.form['confirm'] = 'bleh'
        with mail_ext.record_messages() as outbox:
            resp = resp.form.submit()

        assert resp.status_code == 302
        assert resp.location.endswith('/users')
        assert resp.flashes == [('success', 'Successfully created User')]
        assert len(outbox) == 0

        # be sure the password is stored. Force-verify the email so we can continue
        user = self.user_ent.get_by(email='foobar@baz.com')
        user.is_verified = True
        db.session.commit()
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/login')
        resp.form['login_id'] = 'foobar@baz.com'
        resp.form['password'] = 'bleh'
        resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/')

    @mock.patch.dict('flask.current_app.config', {'KEGAUTH_EMAIL_OPS_ENABLED': False})
    def test_edit_no_email_same_password(self):
        user = self.user_ent.fake()
        resp = self.client.get('/users/{}/edit'.format(user.id))

        resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/users')
        assert resp.flashes == [('success', 'Successfully modified User')]

        # be sure the password hasn't changed
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/login')
        resp.form['login_id'] = user.email
        resp.form['password'] = user._plaintext_pass
        resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/')

    @mock.patch.dict('flask.current_app.config', {'KEGAUTH_EMAIL_OPS_ENABLED': False})
    def test_edit_no_email_reset_password(self):
        user = self.user_ent.fake(password='foobar')
        resp = self.client.get('/users/{}/edit'.format(user.id))

        resp.form['reset_password'] = 'bleh'
        resp.form['confirm'] = 'bleh'
        resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/users')
        assert resp.flashes == [('success', 'Successfully modified User')]

        # be sure the password hasn't changed
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/login')
        resp.form['login_id'] = user.email
        resp.form['password'] = 'bleh'
        resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/')

    def test_add_with_session_key(self):
        resp = self.client.get('/users/add?session_key=foo')
        assert resp.pyquery('a.cancel').attr('href').endswith('/users?session_key=foo')

        resp.form['email'] = 'abc@example.session.com'

        resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/users?session_key=foo')

    @mock.patch('flask.render_template', autospec=True, spec_set=True)
    def test_add_template_args(self, m_render):
        m_render.return_value = 'foo'
        self.client.get('/users/add')
        assert 'a_template_variable' in m_render.call_args[1]

    def test_edit(self):
        user_edit = ents.User.fake(
            groups=[ents.Group.fake()],
            bundles=[ents.Bundle.fake()],
            permissions=[ents.Permission.fake()],
        )

        resp = self.client.get('/users/{}/edit'.format(user_edit.id))
        assert resp.pyquery('#page-content')('h1').eq(0).text() == 'Edit User'
        assert resp.form['email'].value == user_edit.email
        assert resp.form['group_ids'].value == [str(obj.id) for obj in user_edit.groups]
        assert resp.form['bundle_ids'].value == [str(obj.id) for obj in user_edit.bundles]
        all_permissions = [p.description or p.token for p in ents.Permission.query.all()]
        user_permissions = [p.description or p.token for p in user_edit.permissions]
        listed_permissions = resp.pyquery('#permission_ids')('.custom-checkbox')
        assert len(listed_permissions) == len(all_permissions)
        for permission_list_item in listed_permissions:
            assert permission_list_item.find('label').text in all_permissions
            if permission_list_item.find('label').text in user_permissions:
                assert 'checked' in permission_list_item.find('input').attrib
        resp.form['email'] = 'foo@bar.baz'
        resp = resp.form.submit()

        assert resp.status_code == 302
        assert resp.location.endswith('/users')
        assert resp.flashes == [('success', 'Successfully modified User')]
        assert self.user_ent.get_by(email='foo@bar.baz')

    def test_edit_disallowed_by_permission(self):
        user = ents.User.fake()
        client = AuthTestApp(flask.current_app, user=user)
        client.get('/users/99999/edit', status=403)

    def test_edit_with_session_key(self):
        user_edit = ents.User.fake()

        resp = self.client.get('/users/{}/edit?session_key=foo'.format(user_edit.id))
        assert resp.pyquery('a.cancel').attr('href').endswith('/users?session_key=foo')

        resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/users?session_key=foo')

    def test_edit_triggers_user_session_key_refresh(self):
        target_user = ents.User.fake(permissions='auth-manage')
        target_user_client = AuthTestApp(flask.current_app, user=target_user)
        new_perm = ents.Permission.fake()
        original_session_key = target_user.session_key

        # target user has matching session key and rights to page
        target_user_client.get('/users', status=200)

        resp = self.client.get('/users/{}/edit'.format(target_user.id))
        resp.form['permission_ids'] = [new_perm.id]
        resp = resp.form.submit()

        db.session.expire(target_user)
        assert target_user.session_key != original_session_key

        # target user should need to log in now
        assert '/login' in target_user_client.get('/users', status=302).location

    def test_not_found(self):
        self.client.get('/users/999999/edit', status=404)
        self.client.get('/users/999999/delete', status=404)

    @pytest.mark.parametrize('action', [
        'add', 'edit', 'delete', 'list'
    ])
    def test_alternate_permissions(self, action):
        # patch in separate permissions for add/edit/view/delete
        actions = {'add', 'edit', 'delete', 'list'}
        ents.Permission.fake(token='permission1')

        user_edit = ents.User.fake()
        user_delete = ents.User.fake()

        def url(url_action):
            if url_action == 'list':
                return '/users'
            if url_action == 'edit':
                return '/users/{}/edit'.format(user_edit.id)
            if url_action == 'delete':
                return '/users/{}/delete'.format(user_delete.id)
            return '/users/add'

        with mock.patch.dict(
            flask.current_app.view_functions[
                'auth.user:{}'.format(action)
            ].view_class.permissions,
            {action: 'permission1'}
        ):
            user = ents.User.fake(permissions='auth-manage')
            client = AuthTestApp(flask.current_app, user=user)
            client.get(url(action), status=403)
            for url_action in actions.difference({action}):
                print(url_action, url(url_action))
                client.get(url(url_action))

            user = ents.User.fake(permissions=('auth-manage', 'permission1'))
            client = AuthTestApp(flask.current_app, user=user)
            client.get(url(action))

    def test_delete(self):
        user_delete_id = ents.User.fake().id

        resp = self.client.get('/users/{}/delete'.format(user_delete_id))

        assert resp.status_code == 302
        assert resp.location.endswith('/users')
        assert resp.flashes == [('success', 'Successfully removed User')]

        assert not self.user_ent.get(user_delete_id)

    def test_delete_disallowed_by_action_init(self):
        user_delete_id = ents.User.fake().id

        def init_object_delete(self):
            flask.abort(403)

        from keg_auth import views
        views.User.init_object_delete = init_object_delete

        try:
            self.client.get('/users/{}/delete'.format(user_delete_id), status=403)
        finally:
            del views.User.init_object_delete

        self.client.get('/users/{}/delete'.format(user_delete_id), status=302)

    def test_delete_disallowed_by_permission(self):
        user = ents.User.fake()
        client = AuthTestApp(flask.current_app, user=user)
        client.get('/users/99999/delete', status=403)

    def test_delete_with_session_key(self):
        user_delete = ents.User.fake()

        resp = self.client.get('/users/{}/delete?session_key=foo'.format(user_delete.id))

        assert resp.status_code == 302
        assert resp.location.endswith('/users?session_key=foo')

    def test_delete_myself_fails(self):
        resp = self.client.get('/users/{}/delete'.format(self.current_user.id))

        assert resp.status_code == 302
        assert resp.location.endswith('/users')
        assert resp.flashes == [('warning',
                                 'Unable to delete User. It may be referenced by other items.')]

    @mock.patch('keg_elements.db.mixins.db.session.delete', autospec=True, spec_set=True)
    def test_delete_failed(self, m_delete):
        m_delete.side_effect = sa.exc.IntegrityError(None, None, None)
        user_delete_id = ents.User.fake().id

        resp = self.client.get('/users/{}/delete'.format(user_delete_id))

        assert resp.status_code == 302
        assert resp.location.endswith('/users')
        assert resp.flashes == [('warning',
                                 'Unable to delete User. It may be referenced by other items.')]

        assert self.user_ent.get(user_delete_id)

    def test_list(self):
        resp = self.client.get('/users?op(username)=eq&v1(username)=' + self.current_user.email)
        assert resp.pyquery('.grid-header-add-link a').attr('href').startswith('/users/add')
        assert resp.pyquery('.datagrid table.records thead th').eq(1).text() == 'User ID'
        assert resp.pyquery('.datagrid table.records tbody td').eq(1).text() == self.current_user.email  # noqa
        assert resp.pyquery('#page-content')('h1').eq(0).text() == 'Users'
        assert resp.pyquery('.grid-header-add-link').eq(0).text() == 'Create User'

    @mock.patch('flask.render_template', autospec=True, spec_set=True)
    def test_list_template_args(self, m_render):
        m_render.return_value = 'foo'
        self.client.get('/users')
        assert 'a_template_variable' in m_render.call_args[1]

    def test_list_alternate_ident_field(self):
        # with the mock here, authentication/authorization will also be happening on the alternate
        # class, which doesn't have relationships like permissions. So, to make this easy, make it
        # a superuser.
        user = ents.UserNoEmail.fake(is_superuser=True)
        client = AuthTestApp(flask.current_app, user=user)
        with mock.patch('keg_auth_ta.extensions.auth_entity_registry._user_cls', ents.UserNoEmail):
            resp = client.get('/users?op(username)=eq&v1(username)=' + user.username)
            assert resp.pyquery('.datagrid table.records thead th').eq(1).text() == 'User ID'
            assert resp.pyquery('.datagrid table.records tbody td').eq(1).text() == user.username

    @mock.patch('keg_auth_ta.extensions.Grid.manager.args_loaders',
                [RequestArgsLoader, RequestFormLoader])
    def test_list_post(self):
        resp = self.client.get('/users')
        form = resp.forms[0]
        assert form.method == 'post'
        resp = form.submit()
        assert resp.status_code == 302
        assert '/users' in resp.location

    @mock.patch('keg_auth_ta.extensions.Grid.manager.args_loaders', [RequestArgsLoader])
    def test_list_post_no_form_loader(self):
        self.client.post('/users', status=400)

    def test_list_export(self):
        ents.User.fake()
        resp = self.client.get('/users?export_to=xlsx')
        assert resp.content_type == 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'  # noqa


class TestGroupCrud(ViewTestBase):
    permissions = 'auth-manage'

    def test_add(self):
        perm_approve = ents.Permission.fake()
        ents.Permission.fake()
        bundle_approve = ents.Bundle.fake()
        ents.Bundle.fake()

        resp = self.client.get('/groups/add')

        assert resp.form['name'].value == ''

        resp.form['name'] = 'test adding a group'
        resp.form['permission_ids'] = [perm_approve.id]
        resp.form['bundle_ids'] = [bundle_approve.id]

        resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/groups')
        assert resp.flashes == [('success', 'Successfully created Group')]

        group = ents.Group.get_by(name='test adding a group')
        assert group.permissions == [perm_approve]
        assert group.bundles == [bundle_approve]

    def test_edit(self):
        group_edit = ents.Group.fake(bundles=[ents.Bundle.fake()],
                                     permissions=[ents.Permission.fake()])

        resp = self.client.get('/groups/{}/edit'.format(group_edit.id))
        assert resp.form['name'].value == group_edit.name
        assert resp.form['bundle_ids'].value == [str(obj.id) for obj in group_edit.bundles]
        all_permissions = [p.description or p.token for p in ents.Permission.query.all()]
        user_permissions = [p.description or p.token for p in group_edit.permissions]
        listed_permissions = resp.pyquery('#permission_ids')('.custom-checkbox')
        assert len(listed_permissions) == len(all_permissions)
        for permission_list_item in listed_permissions:
            assert permission_list_item.find('label').text in all_permissions
            if permission_list_item.find('label').text in user_permissions:
                assert 'checked' in permission_list_item.find('input').attrib
        resp.form['name'] = 'test editing a group'
        resp = resp.form.submit()

        assert resp.status_code == 302
        assert resp.location.endswith('/groups')
        assert resp.flashes == [('success', 'Successfully modified Group')]
        assert ents.Group.get_by(name='test editing a group')

    def test_not_found(self):
        self.client.get('/groups/999999/edit', status=404)
        self.client.get('/groups/999999/delete', status=404)

    @pytest.mark.parametrize('action', [
        'add', 'edit', 'delete', 'list'
    ])
    def test_alternate_permissions(self, action):
        # patch in separate permissions for add/edit/view/delete
        actions = {'add', 'edit', 'delete', 'list'}
        ents.Permission.fake(token='permission1')

        group_edit = ents.Group.fake()
        group_delete = ents.Group.fake()

        def url(url_action):
            if url_action == 'list':
                return '/groups'
            if url_action == 'edit':
                return '/groups/{}/edit'.format(group_edit.id)
            if url_action == 'delete':
                return '/groups/{}/delete'.format(group_delete.id)
            return '/groups/add'

        with mock.patch.dict(
            flask.current_app.view_functions[
                'auth.group:{}'.format(action)
            ].view_class.permissions,
            {action: 'permission1'}
        ):
            user = ents.User.fake(permissions='auth-manage')
            client = AuthTestApp(flask.current_app, user=user)
            client.get(url(action), status=403)
            for url_action in actions.difference({action}):
                print(url_action, url(url_action))
                client.get(url(url_action))

            user = ents.User.fake(permissions=('auth-manage', 'permission1'))
            client = AuthTestApp(flask.current_app, user=user)
            client.get(url(action))

    def test_delete(self):
        group_delete_id = ents.Group.fake().id

        resp = self.client.get('/groups/{}/delete'.format(group_delete_id))

        assert resp.status_code == 302
        assert resp.location.endswith('/groups')
        assert resp.flashes == [('success', 'Successfully removed Group')]

        assert not ents.Group.get(group_delete_id)

    @mock.patch('keg_elements.db.mixins.db.session.delete', autospec=True, spec_set=True)
    def test_delete_failed(self, m_delete):
        m_delete.side_effect = sa.exc.IntegrityError(None, None, None)
        group_delete_id = ents.Group.fake().id

        resp = self.client.get('/groups/{}/delete'.format(group_delete_id))

        assert resp.status_code == 302
        assert resp.location.endswith('/groups')
        assert resp.flashes == [('warning',
                                 'Unable to delete Group. It may be referenced by other items.')]

        assert ents.Group.get(group_delete_id)

    def test_list(self):
        ents.Group.fake()
        resp = self.client.get('/groups')
        assert resp.pyquery('.grid-header-add-link a').attr('href').startswith('/groups/add')
        assert 'datagrid' in resp

    def test_list_export(self):
        ents.Group.fake()
        resp = self.client.get('/groups?export_to=xlsx')
        assert resp.content_type == 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'  # noqa


class TestBundleCrud(ViewTestBase):
    permissions = 'auth-manage'

    @classmethod
    def setup_class(cls):
        super().setup_class()
        ents.Group.delete_cascaded()
        ents.Bundle.delete_cascaded()

    def test_add(self):
        perm_approve = ents.Permission.fake()
        ents.Permission.fake()

        resp = self.client.get('/bundles/add')

        assert resp.form['name'].value == ''

        resp.form['name'] = 'test adding a bundle'
        resp.form['permission_ids'] = [perm_approve.id]

        resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/bundles')
        assert resp.flashes == [('success', 'Successfully created Bundle')]

        bundle = ents.Bundle.get_by(name='test adding a bundle')
        assert bundle.permissions == [perm_approve]

    def test_edit(self):
        bundle_edit = ents.Bundle.fake(permissions=[ents.Permission.fake()])

        resp = self.client.get('/bundles/{}/edit'.format(bundle_edit.id))
        assert resp.form['name'].value == bundle_edit.name
        all_permissions = [p.description or p.token for p in ents.Permission.query.all()]
        user_permissions = [p.description or p.token for p in bundle_edit.permissions]
        listed_permissions = resp.pyquery('#permission_ids')('.custom-checkbox')
        assert len(listed_permissions) == len(all_permissions)
        for permission_list_item in listed_permissions:
            assert permission_list_item.find('label').text in all_permissions
            if permission_list_item.find('label').text in user_permissions:
                assert 'checked' in permission_list_item.find('input').attrib
        resp.form['name'] = 'test editing a bundle'
        resp = resp.form.submit()

        assert resp.status_code == 302
        assert resp.location.endswith('/bundles')
        assert resp.flashes == [('success', 'Successfully modified Bundle')]
        assert ents.Bundle.get_by(name='test editing a bundle')

    def test_not_found(self):
        self.client.get('/bundles/999999/edit', status=404)
        self.client.get('/bundles/999999/delete', status=404)

    @pytest.mark.parametrize('action', [
        'add', 'edit', 'delete', 'list'
    ])
    def test_alternate_permissions(self, action):
        # patch in separate permissions for add/edit/view/delete
        actions = {'add', 'edit', 'delete', 'list'}
        ents.Permission.fake(token='permission1')

        bundle_edit = ents.Bundle.fake()
        bundle_delete = ents.Bundle.fake()

        def url(url_action):
            if url_action == 'list':
                return '/bundles'
            if url_action == 'edit':
                return '/bundles/{}/edit'.format(bundle_edit.id)
            if url_action == 'delete':
                return '/bundles/{}/delete'.format(bundle_delete.id)
            return '/bundles/add'

        with mock.patch.dict(
            flask.current_app.view_functions[
                'auth.bundle:{}'.format(action)
            ].view_class.permissions,
            {action: 'permission1'}
        ):
            user = ents.User.fake(permissions='auth-manage')
            client = AuthTestApp(flask.current_app, user=user)
            client.get(url(action), status=403)
            for url_action in actions.difference({action}):
                print(url_action, url(url_action))
                client.get(url(url_action))

            user = ents.User.fake(permissions=('auth-manage', 'permission1'))
            client = AuthTestApp(flask.current_app, user=user)
            client.get(url(action))

    def test_delete(self):
        bundle_delete_id = ents.Bundle.fake().id

        resp = self.client.get('/bundles/{}/delete'.format(bundle_delete_id))

        assert resp.status_code == 302
        assert resp.location.endswith('/bundles')
        assert resp.flashes == [('success', 'Successfully removed Bundle')]

        assert not ents.Bundle.get(bundle_delete_id)

    @mock.patch('keg_elements.db.mixins.db.session.delete', autospec=True, spec_set=True)
    def test_delete_failed(self, m_delete):
        m_delete.side_effect = sa.exc.IntegrityError(None, None, None)
        bundle_delete_id = ents.Bundle.fake().id

        resp = self.client.get('/bundles/{}/delete'.format(bundle_delete_id))

        assert resp.status_code == 302
        assert resp.location.endswith('/bundles')
        assert resp.flashes == [('warning',
                                 'Unable to delete Bundle. It may be referenced by other items.')]

        assert ents.Bundle.get(bundle_delete_id)

    def test_list(self):
        obj = ents.Bundle.fake()
        resp = self.client.get('/bundles')
        assert resp.pyquery('.grid-header-add-link a').attr('href').startswith('/bundles/add')
        assert 'datagrid' in resp
        assert obj.name in resp

    def test_list_export(self):
        ents.Bundle.fake()
        resp = self.client.get('/bundles?export_to=xlsx')
        assert resp.content_type == 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'  # noqa


class TestPermissionsView(ViewTestBase):
    permissions = 'auth-manage'

    def test_list(self):
        ents.Permission.fake()
        resp = self.client.get('/permissions')
        assert not resp.pyquery('.grid-header-add-link')
        assert 'datagrid' in resp

    def test_list_export(self):
        ents.Permission.fake()
        resp = self.client.get('/permissions?export_to=xlsx')
        assert resp.content_type == 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'  # noqa


class TestGetCurrentUser:
    def test_no_user_returned(self):
        with flask.current_app.test_request_context():
            assert get_current_user() is None

    def test_flask_login_user_returned(self):
        with flask.current_app.test_request_context():
            user = ents.User.fake()
            flask_login.login_user(user)
            assert get_current_user().id == user.id

    def test_request_loader_user_returned(self, app_context):
        with flask.current_app.test_request_context():
            user = ents.User.fake()
            jwt_auth = flask.current_app.auth_manager.get_request_loader('jwt')
            token = jwt_auth.create_access_token(user)
            flask.request.headers = MultiDict([
                ('Authorization', 'Bearer {}'.format(token)),
            ])

            assert get_current_user().id == user.id


class TestViewTestBase:
    def test_permission_validated_from_tuple(self):
        class Foo(ViewTestBase):
            permissions = 'foo'

        with mock.patch(
            'keg.current_app.auth_manager.permissions', [('foo', 'bar'), ('baz', 'bam')]
        ):
            Foo.setup_class()

    def test_permission_validated_from_string(self):
        class Foo(ViewTestBase):
            permissions = 'foo'

        with mock.patch(
            'flask.current_app.auth_manager.permissions', ['foo', 'baz']
        ):
            Foo.setup_class()

    def test_permission_invalid(self):
        class Foo(ViewTestBase):
            permissions = 'bar'

        with mock.patch(
            'flask.current_app.auth_manager.permissions', ['foo', 'baz']
        ):
            with pytest.raises(
                Exception, match=r"permission\(s\) \{'bar'\} not specified in the auth manager"
            ):
                Foo.setup_class()

    def test_multiple_permissions_validated(self):
        class Foo(ViewTestBase):
            permissions = 'foo', 'baz'

        with mock.patch(
            'flask.current_app.auth_manager.permissions', ['foo', 'baz']
        ):
            Foo.setup_class()


@pytest.fixture
def auth_user():
    user = ents.User.fake()
    yield user


@pytest.fixture
def redirect_authenticator():
    authenticator = RedirectAuthenticator(flask.current_app)
    with mock.patch(
        'flask.current_app.auth_manager.login_authenticator',
        authenticator,
    ):
        yield authenticator


class TestRedirectAuthenticator(ViewTestBase):
    @mock.patch.dict('flask.current_app.config', {'KEGAUTH_REDIRECT_LOGIN_TARGET': '/l2'})
    def test_login(self, redirect_authenticator):
        resp = self.client.get('/login', status=302)
        assert resp.location.endswith('/l2')

    def test_login_no_target(self, redirect_authenticator):
        with pytest.raises(Exception, match='KEGAUTH_REDIRECT_LOGIN_TARGET not set.*'):
            self.client.get('/login')

    @mock.patch.dict('flask.current_app.config', {'KEGAUTH_REDIRECT_LOGIN_TARGET': 'https://l2'})
    def test_login_target_absolute(self, redirect_authenticator):
        with pytest.raises(Exception, match='KEGAUTH_REDIRECT_LOGIN_TARGET not set.*'):
            self.client.get('/login')

    def test_not_found_targets(self, redirect_authenticator):
        self.client.get('/forgot-password', status=404)
        self.client.get('/reset-password', status=404)
        self.client.get('/verify-account', status=404)

    def test_logout(self, redirect_authenticator):
        resp = self.client.get('/logout', status=302)
        assert resp.flashes == [('success', 'You have been logged out.')]


@pytest.fixture
def oauth_client():
    with mock.patch.dict(
        'flask.current_app.config',
        {'KEGAUTH_OAUTH_PROFILES': [oauth_profile()]}
    ):
        authenticator = OAuthAuthenticator(flask.current_app)
        flask.current_app.auth_manager.oauth_authenticator = authenticator
        client = flask.current_app.auth_manager.oauth.create_client('google')
        yield client


class TestOAuthLogin:
    def test_profile_not_found(self, oauth_client):
        client = AuthTestApp(flask.current_app)
        client.get('/login/foo', status=404)

    def test_authenticated(self, oauth_client):
        client = AuthTestApp(flask.current_app)
        with mock.patch.object(oauth_client, 'authorize_redirect') as m_redirect:
            m_redirect.return_value = 'foo'
            resp = client.get('/login/google')
            assert resp.body == b'foo'
            m_redirect.assert_called_once_with('http://keg.example.com/oauth-authorize/google')


class TestOAuthAuthorize:
    def setup_method(self):
        ents.User.delete_cascaded()

    def test_profile_not_found(self, oauth_client):
        client = AuthTestApp(flask.current_app)
        client.get('/oauth-authorize/foo', status=404)

    def test_user_found(self, oauth_client):
        auth_user = ents.User.fake(email='foo@mycompany.biz')
        client = AuthTestApp(flask.current_app)
        with mock.patch.object(oauth_client, 'authorize_access_token') as m_auth:
            m_auth.return_value = {'userinfo': {'email': auth_user.email}}
            resp = client.get('/oauth-authorize/google', status=302)
            assert resp.location == '/'
            assert resp.flashes == [('success', 'Login successful.')]

    def test_user_found_alt_token_path(self, oauth_client):
        auth_user = ents.User.fake(email='foo@mycompany.biz')
        client = AuthTestApp(flask.current_app)
        with mock.patch.object(oauth_client, 'authorize_access_token') as m_auth:
            m_auth.return_value = {}
            with mock.patch.object(oauth_client, 'userinfo') as m_info:
                m_info.return_value = {'email': auth_user.email}
                resp = client.get('/oauth-authorize/google', status=302)
            assert resp.location == '/'
            assert resp.flashes == [('success', 'Login successful.')]

    def test_not_exists(self, oauth_client):
        client = AuthTestApp(flask.current_app)
        with mock.patch.object(oauth_client, 'authorize_access_token') as m_auth:
            m_auth.return_value = {'userinfo': {'email': 'foo@mycompany.biz'}}
            resp = client.get('/oauth-authorize/google')
            assert 'No user account matches' in resp

    def test_user_inactive(self, oauth_client):
        auth_user = ents.User.fake(email='foo@mycompany.biz', is_enabled=False)
        client = AuthTestApp(flask.current_app)
        with mock.patch.object(oauth_client, 'authorize_access_token') as m_auth:
            m_auth.return_value = {'userinfo': {'email': auth_user.email}}
            resp = client.get('/oauth-authorize/google')
            assert 'has been disabled' in resp



from collections import namedtuple

from blazeutils.strings import normalizews
import flask
from markdown_it import MarkdownIt

try:
    import flask_mail
except ImportError:
    pass

MailParts = namedtuple('MailParts', 'subject text html')


def mail_template(template_name_or_list, **kwargs):
    multi_part_content = flask.render_template(template_name_or_list, **kwargs)
    parts = multi_part_content.split('---multi-part:xFE+Ab7j+w,mdIL%---')
    subject, markdown = map(lambda p: p.strip(), parts)

    return MailParts(
        normalizews(subject),
        markdown,
        MarkdownIt().render(markdown)
    )


class AuthMailManager(object):
    """Manager to handle sending auth-related mail via a flask-mail extension.

    Pass the mail extension in the constructor.
    """
    reset_password_templates = ('mail/reset-password.j2', 'keg-auth/reset-password-mail.j2')
    new_user_templates = ('mail/new-user.j2', 'keg-auth/new-user-mail.j2')

    def __init__(self, mail_ext):
        self.mail_ext = mail_ext

    def reset_password_message(self, user):
        """Creates message from `reset_password_template` with the given user."""
        parts = mail_template(self.reset_password_templates, user=user)

        return flask_mail.Message(parts.subject, [user.email], parts.text, parts.html)

    def send_reset_password(self, user):
        """Send password reset email from `reset_password_message` with the given user."""
        msg = self.reset_password_message(user)
        self.mail_ext.send(msg)

    def new_user_message(self, user):
        """Creates message from `new_user_templates` with the given user."""
        parts = mail_template(self.new_user_templates, user=user)

        return flask_mail.Message(parts.subject, [user.email], parts.text, parts.html)

    def send_new_user(self, user):
        """Send account creation email from `new_user_message` with the given user."""
        msg = self.new_user_message(user)
        self.mail_ext.send(msg)

    def verify_account_url(self, user):
        """Returns URL to use in mail for account verification for the given user."""
        return flask.current_app.auth_manager.url_for(
            'verify-account', user_id=user.id, token=user._token_plain, _external=True)

    def reset_password_url(self, user):
        """Returns URL to use in mail for password reset for the given user."""
        return flask.current_app.auth_manager.url_for(
            'reset-password', user_id=user.id, token=user._token_plain, _external=True)



import warnings

from keg.testing import ContextManager
import pytest

from keg_auth_ta.app import KegAuthTestApp

# Because our tests are in keg_auth.testing, which isn't a test module, pytest won't rewrite the
# assertions by default.
pytest.register_assert_rewrite('keg_auth.testing')

warnings.filterwarnings(
    'error', r"Unicode type received non-unicode bind param value",
    module='sqlalchemy.sql.sqltypes'
)


def pytest_configure(config):
    KegAuthTestApp.testing_prep()


@pytest.fixture(scope='class', autouse=True)
def auto_app_context():
    with ContextManager.get_for(KegAuthTestApp).app.app_context():
        yield



from blazeutils.strings import randchars
from keg import config


class DefaultProfile(object):
    SECRET_KEY = randchars()

    KEG_BASE_TEMPLATE = 'base-page.html'
    KEGAUTH_TEMPLATE_TITLE_VAR = 'title'

    # These three just get rid of warnings on the console.
    KEG_KEYRING_ENABLE = False
    SQLALCHEMY_DATABASE_URI = 'postgresql://postgres@localhost/postgres'
    SQLALCHEMY_TRACK_MODIFICATIONS = False

    SITE_NAME = 'Keg Auth Demo'
    SITE_ABBR = 'KA Demo'

    AUTO_EXPAND_MENU = True


class TestProfile(object):
    # Make tests faster
    PASSLIB_CRYPTCONTEXT_KWARGS = dict(schemes=['plaintext'])

    MAIL_DEFAULT_SENDER = 'sender@example.com'

    # These settings reflect what is needed in CI & when using docker compose.
    # Use keg_auth_ta-config.py to override if needed.
    SQLALCHEMY_DATABASE_URI = 'postgresql://postgres@localhost/kegauth_tests'


class TestProfileUserArgs(config.TestProfile, TestProfile):
    KEGAUTH_CLI_USER_ARGS = ['name', 'email']



from keg.signals import app_ready

from keg_auth import NavItem, NavURL

from keg_auth_ta.cli import auth_cli_extensions


@app_ready.connect
def init_app_cli(app):
    auth_cli_extensions(app)


@app_ready.connect
def init_navigation(app):
    app.auth_manager.add_navigation_menu(
        'main',
        NavItem(
            NavItem('Home', NavURL('public.home')),
            NavItem(
                'Sub-Menu',
                NavItem('User Manage', NavURL('private.secret2')),
                NavItem('Secret View', NavURL('private.secret_nested'), class_='my-link-class'),
                class_='my-group-class'
            ),
            NavItem(
                'Menu-Group',
                NavItem('User Manage 2', NavURL('auth.user:list')),
                NavItem('User Manage 3', NavURL('auth.user:list'), icon_class='fas fa-ad'),
                nav_group='auth',
                icon_class='fas fa-bomb'
            ),
        )
    )



from keg_auth.extensions import lazy_gettext as _


class PermissionCondition(object):
    """Basic permission condition that will return True/False upon user access check.

    One or more permissions or conditions must be provided to the constructor.
    """
    def __init__(self, *conditions):
        if len(conditions) < 1:
            raise ValueError(_('At least one permission or condition is required'))
        self.conditions = conditions

    @staticmethod
    def _check_condition(condition, user):
        if isinstance(condition, PermissionCondition):
            return condition.check(user)
        if callable(condition):
            return condition(user)
        if hasattr(user, 'has_all_permissions'):
            return user.has_all_permissions(condition)

        # probably an anonymous user in the session after logout
        return False

    def check(self, user):
        raise NotImplementedError  # pragma: no cover


class AllCondition(PermissionCondition):
    """Condition requiring all contained permissions/conditions to be satisfied.

    Rules governing the contained permission/conditions:
    - Not all conditions are guaranteed to be checked. Checking will exit on the first failure.
    - Callable conditions are expected to take a `user` argument.
    - Permission token conditions are run if the user has a `has_all_permissions` method
    (default case).
    """
    def check(self, user):
        for cond in self.conditions:
            if not self._check_condition(cond, user):
                return False
        return True


class AnyCondition(PermissionCondition):
    """Condition requiring only one of the contained permissions/conditions to be satisfied.

    Rules governing the contained permission/conditions:
    - Not all conditions are guaranteed to be checked. Checking will exit on the first success.
    - Callable conditions are expected to take a `user` argument.
    - Permission token conditions are run if the user has a `has_all_permissions` method
    (default case).
    """
    def check(self, user):
        for cond in self.conditions:
            if self._check_condition(cond, user):
                return True
        return False


def has_permissions(condition, user):
    """Check a user against a single condition/permission."""
    if condition is None:
        return True
    return PermissionCondition._check_condition(condition, user)


has_all = AllCondition
has_any = AnyCondition



from contextlib import contextmanager


@contextmanager
def listen_to(signal):
    ''' Context Manager that listens to signals and records emissions
    Example:
    with listen_to(user_logged_in) as listener:
        login_user(user)
        # Assert that a single emittance of the specific args was seen.
        listener.assert_heard_one(app, user=user))
        # Of course, you can always just look at the list yourself
        self.assertEqual(1, len(listener.heard))
    '''
    class _SignalsCaught(object):
        def __init__(self):
            self.heard = []

        def add(self, *args, **kwargs):
            ''' The actual handler of the signal. '''
            self.heard.append((args, kwargs))

        def assert_heard_one(self, *args, **kwargs):
            ''' The signal fired once, and with the arguments given '''
            if len(self.heard) == 0:
                raise AssertionError('No signals were fired')
            elif len(self.heard) > 1:
                msg = '{0} signals were fired'.format(len(self.heard))
                raise AssertionError(msg)
            elif self.heard[0] != (args, kwargs):
                msg = 'One signal was heard, but with incorrect arguments: '\
                    'Got ({0}) expected ({1}, {2})'
                raise AssertionError(msg.format(self.heard[0], args, kwargs))

        def assert_heard_none(self, *args, **kwargs):
            ''' The signal fired no times '''
            if len(self.heard) >= 1:
                msg = '{0} signals were fired'.format(len(self.heard))
                raise AssertionError(msg)

    results = _SignalsCaught()
    signal.connect(results.add)

    try:
        yield results
    finally:
        signal.disconnect(results.add)


def oauth_profile(**kwargs):
    base = {
        'domain_filter': 'mycompany.biz',
        'id_field': 'email',
        'oauth_client_kwargs': {
            'name': 'google',
            'client_id': 'my-google-client',
            'client_secret': 'super-secret',
            'server_metadata_url': 'http://mysite/openid-configuration',
            'client_kwargs': {'scope': 'openid email'},
        }
    }
    base.update(**kwargs)
    return base



from pathlib import Path
import subprocess
import sys

import click


def auth_cli_extensions(app):
    @app.auth_manager.cli_group.command('command-extension')
    def command_extension():
        """ Demonstrate extending the auth CLI group.

            The AuthManager holds the CLI group created for user management, but extending that
            group is not as simple as it could be. To have access to the auth manager, we must
            have an active app context, so the command extension gets wrapped in a function,
            which is called via an event handler when the app is ready.
        """
        click.echo('verified')

    @app.cli.command('verify-translations', help='Verifies all strings marked for translation')
    def verify_translations():
        from pathlib import Path
        from morphi.messages.validation import check_translations

        root_path = Path(__file__).resolve().parent.parent
        check_translations(
            root_path,
            'keg_auth',
            ignored_strings = {
                'Email',
            }
        )



DEFAULT_PROFILE = 'DevProfile'


class DevProfile(object):
    MAIL_DEFAULT_SENDER = ''
    SERVER_NAME = 'localhost:5000'
    # SQLALCHEMY_DATABASE_URI = 'postgresql://postgres@localhost/postgres'


class TestProfile(object):
    # SQLALCHEMY_DATABASE_URI = 'postgresql://postgres@localhost/kegauth_tests'
    pass



import datetime as dt

import arrow
import flask
import keg
from keg_elements.forms import Form, ModelForm, FieldMeta, MultiCheckboxField
from keg_elements.forms.validators import ValidateUnique
from sqlalchemy.sql.functions import coalesce
from sqlalchemy_utils import EmailType
from wtforms.fields import (
    BooleanField,
    DateField,
    HiddenField,
    PasswordField,
    StringField,
    SelectMultipleField)
from wtforms import ValidationError, validators
from wtforms_components.widgets import EmailInput

from keg_auth.extensions import lazy_gettext as _
from keg_auth.libs import get_domain_from_email
from keg_auth.libs.templates import link_to
from keg_auth.model import get_username_key


def login_form():
    """Returns a Login form class that handles username options."""
    login_id_label = _(u'User ID')
    login_id_validators = [validators.DataRequired()]

    if isinstance(flask.current_app.auth_manager.entity_registry.user_cls.username.type, EmailType):
        login_id_label = _(u'Email')
        login_id_validators.append(validators.Email())

    class Login(Form):
        next = HiddenField()

        login_id = StringField(login_id_label, validators=login_id_validators)
        password = PasswordField(_('Password'), validators=[
            validators.DataRequired(),
        ])

    return Login


class ForgotPassword(Form):
    """Returns a form to capture email for password reset."""
    email = StringField(_(u'Email'), validators=[
        validators.DataRequired(),
        validators.Email(),
    ])


class SetPassword(Form):
    """Returns a form to capture password/confirmation and apply password policy."""
    password = PasswordField(_('New Password'), validators=[
        validators.DataRequired(),
        validators.EqualTo('confirm', message=_('Passwords must match'))
    ])
    confirm = PasswordField(_('Confirm Password'))

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop('user')
        super().__init__(*args, **kwargs)

        auth_manager = keg.current_app.auth_manager
        password_policy = auth_manager.password_policy_cls()
        self.password.validators = [*self.password.validators, *password_policy.form_validators()]


def get_permission_options():
    perm_cls = flask.current_app.auth_manager.entity_registry.permission_cls
    query = perm_cls.query.with_entities(
        perm_cls.id, coalesce(perm_cls.description, perm_cls.token).label('desc')
    ).order_by('desc').all()
    return [(str(perm.id), perm.desc) for perm in query]


def get_bundle_options():
    bundle_cls = flask.current_app.auth_manager.entity_registry.bundle_cls
    return [(str(bundle.id), bundle.name) for bundle in bundle_cls.query.order_by('name')]


def get_group_options():
    group_cls = flask.current_app.auth_manager.entity_registry.group_cls
    return [(str(group.id), group.name) for group in group_cls.query.order_by('name')]


def entities_from_ids(cls, ids):
    if not ids:
        return []
    return cls.query.filter(cls.id.in_(ids)).all()


class GroupsMixin(object):
    group_ids = SelectMultipleField('Groups')

    def after_init(self, args, kwargs):
        self.group_ids.choices = get_group_options()
        if kwargs.get('obj') and self.group_ids.raw_data is None:
            self.group_ids.process_data([group.id for group in kwargs['obj'].groups])
        super().after_init(args, kwargs)

    def get_selected_groups(self):
        return entities_from_ids(flask.current_app.auth_manager.entity_registry.group_cls,
                                 self.group_ids.data)


class PermissionsMixin(object):
    permission_ids = MultiCheckboxField(
        'Permissions',
        render_kw={'class': 'list-unstyled'},
    )

    def after_init(self, args, kwargs):
        self.permission_ids.choices = get_permission_options()
        if kwargs.get('obj') and self.permission_ids.raw_data is None:
            self.permission_ids.process_data([perm.id for perm in kwargs['obj'].permissions])
        super().after_init(args, kwargs)

    def get_selected_permissions(self):
        selected_ids = self.permission_ids.data
        return entities_from_ids(flask.current_app.auth_manager.entity_registry.permission_cls,
                                 selected_ids)


class BundlesMixin(object):
    bundle_ids = SelectMultipleField('Bundles')

    def after_init(self, args, kwargs):
        self.bundle_ids.choices = get_bundle_options()
        if kwargs.get('obj') and self.bundle_ids.raw_data is None:
            self.bundle_ids.process_data([bundle.id for bundle in kwargs['obj'].bundles])
        super().after_init(args, kwargs)

    def get_selected_bundles(self):
        return entities_from_ids(flask.current_app.auth_manager.entity_registry.bundle_cls,
                                 self.bundle_ids.data)


class _ValidatePasswordRequired(object):
    def __call__(self, form, field):
        if not form.obj and not field.data:
            raise ValidationError(_('This field is required.'))
        return True


class _ValidateUsername:
    def __init__(self, username_key=None):
        self.username_key = username_key

    def __call__(self, form, field):
        if form.obj and getattr(form.obj, self.username_key) != field.data:
            original_domain = get_domain_from_email(getattr(form.obj, self.username_key))
            new_domain = get_domain_from_email(field.data)
            is_exclusion = flask.current_app.auth_manager.login_authenticator.is_domain_excluded(
                getattr(form.obj, self.username_key)
            )
            if original_domain and is_exclusion and original_domain != new_domain:
                raise ValidationError(_('Cannot change user domain.'))
        return True


def user_form(config=None, allow_superuser=False, endpoint='',
              fields=['is_enabled', 'disabled_utc']):
    """Returns a form for User CRUD.

    Form is customized depending on the fields and superuser setting passed in."""
    config = config or {}
    user_cls = flask.current_app.auth_manager.entity_registry.user_cls

    # The model can be assumed to have a `username` attribute. However, it may not be settable,
    # depending on whether it is actually a column, or is instead a proxy to another column
    # (e.g. email). So, we will need to grab a key to use for it that matches the data column.
    username_key = get_username_key(flask.current_app.auth_manager.entity_registry.user_cls)

    # create a copy of fields for internal use. In python 2, if we use this as a static method,
    #   the kwarg value would get modified in the wrong scope
    _fields = [username_key] + fields[:]
    if allow_superuser and 'is_superuser' not in _fields:
        _fields.append('is_superuser')

    def html_link(obj):
        return link_to(
            obj.username,
            flask.url_for(endpoint, objid=obj.id)
        )

    def filter_disabled_utc(date):
        if isinstance(date, dt.date):
            date = arrow.get(date)

        return date

    class User(PermissionsMixin, BundlesMixin, GroupsMixin, ModelForm):
        if 'disabled_utc' in _fields:
            disabled_utc = DateField('Disable Date', [validators.Optional()],
                                     filters=[filter_disabled_utc], render_kw={'type': 'date'})

        class Meta:
            model = user_cls
            only = _fields

        class FieldsMeta:
            is_enabled = FieldMeta('Enabled')
            is_superuser = FieldMeta('Superuser')
            __default__ = FieldMeta

        _field_order = tuple(_fields + ['group_ids', 'bundle_ids',
                                        'permission_ids'])

        setattr(FieldsMeta, username_key, FieldMeta(
            extra_validators=[validators.data_required(),
                              ValidateUnique(html_link),
                              _ValidateUsername(username_key)]
        ))

        if isinstance(flask.current_app.auth_manager.entity_registry.user_cls.username.type,
                      EmailType):
            getattr(FieldsMeta, username_key).widget = EmailInput()

        if not config.get('KEGAUTH_EMAIL_OPS_ENABLED'):
            reset_password = PasswordField(_('New Password'), validators=[
                _ValidatePasswordRequired(),
                validators.EqualTo('confirm', message=_('Passwords must match'))
            ])
            confirm = PasswordField(_('Confirm Password'))
            _field_order = _field_order + ('reset_password', 'confirm')
        else:
            # place a Send Welcome field after the initial set of fields
            send_welcome = BooleanField('Send Welcome Email', default=True)
            _field_order = tuple(_fields + ['send_welcome'] + list(_field_order[len(_fields):]))

        def get_object_by_field(self, field):
            return user_cls.get_by(username=field.data)

        @property
        def obj(self):
            return self._obj

        def after_init(self, args, kwargs):
            if kwargs.get('obj') and hasattr(self, 'send_welcome'):
                self.send_welcome = None
                del self._fields['send_welcome']
                self._field_order = tuple(filter(lambda v: v != 'send_welcome', self._field_order))

            return super().after_init(args, kwargs)

    return User


def group_form(endpoint):
    """Returns a form for Group CRUD."""
    group_cls = flask.current_app.auth_manager.entity_registry.group_cls

    def html_link(obj):
        import flask
        return link_to(obj.name, flask.url_for(endpoint, objid=obj.id))

    class Group(PermissionsMixin, BundlesMixin, ModelForm):
        _field_order = ('name', 'bundle_ids', 'permission_ids',)

        class Meta:
            model = group_cls

        class FieldsMeta:
            name = FieldMeta(extra_validators=[ValidateUnique(html_link)])

        def get_object_by_field(self, field):
            return group_cls.get_by(name=field.data)

        @property
        def obj(self):
            return self._obj

    return Group


def bundle_form(endpoint):
    """Returns a form for Bundle CRUD."""
    bundle_cls = flask.current_app.auth_manager.entity_registry.bundle_cls

    def html_link(obj):
        import flask
        return link_to(obj.name, flask.url_for(endpoint, objid=obj.id))

    class Bundle(PermissionsMixin, ModelForm):
        class Meta:
            model = bundle_cls

        class FieldsMeta:
            name = FieldMeta(extra_validators=[ValidateUnique(html_link)])

        def get_object_by_field(self, field):
            return bundle_cls.get_by(name=field.data)

        @property
        def obj(self):
            return self._obj

    return Bundle
